/**
   This class blurs the distinction between a binary tree and
   a binary tree node. This class essentially defines a binary
   tree as a binary tree node. But those two data types are not
   the same thing, so they should not be conflated.

   This class tries to use the null value to represent the empty
   tree. But this doesn't work very well since a null value is
   not a binary tree (if the reference variable b = null, you
   cannot call a method on b, like b.size(), so b cannot be a
   binary tree).
*/
public class BTree
{
   public String element;
   public BTree left;
   public BTree right;

   /**
      Construct a leaf node.
   */
   public BTree(String element)
   {
      this(element, null, null);
   }

   /**
      Construct a BTree with the given binary trees
      as its left and right branches.
   */
   public BTree(String element, BTree left, BTree right)
   {
      if ( null == element )
         throw new NullPointerException("BTree must have a root element");

      this.element = element;
      this.left  = left;
      this.right = right;
   }


   /**
      Determine if this binary tree is a leaf node.
   */
   public boolean isLeaf()
   {
      return null == left
          && null == right;
   }


   /**
      Compute the maximum number of subtrees
      for any node below this node.
   */
   public int degree()
   {
      int thisDegree = 0;
      int leftDegree = 0;
      int rightDegree = 0;

      if ( null != left )
      {
         ++thisDegree;
         leftDegree = left.degree();   // recursion
      }
      if ( null != right )
      {
         ++thisDegree;
         rightDegree = right.degree(); // recursion
      }
      return Math.max(thisDegree, Math.max(leftDegree, rightDegree));
   }


   /**
      Calculate the height of the binary tree rooted at this node.
   */
   public int height()
   {
      int result = 0;
      int leftHeight = 0;
      int rightHeight = 0;

      if ( null != left )
      {
         result = 1;
         leftHeight = left.height();   // recursion
      }
      if ( null != right )
      {
         result = 1;
         rightHeight = right.height(); // recursion
      }

      result += (leftHeight > rightHeight) ? leftHeight : rightHeight;

      return result;
   }


  /**
      Calculate the size of the binary tree rooted at this node.
   */
   public int size()
   {
      int leftSize = 0;
      int rightSize = 0;

      if ( null != left )
      {
         leftSize = left.size();   // recursion
      }
      if ( null != right )
      {
         rightSize = right.size(); // recursion
      }
      return 1 + leftSize + rightSize;
   }


   /**
      Determine if this node is the root of a full binary tree.
   */
   public boolean isFull()
   {
      return ( isLeaf() )
            || ( null != left
              && null != right
              && left.isFull()     // recursion
              && right.isFull() ); // recursion
   }


   /**
      Convert the binary tree rooted at this node into a String.
   */
   @Override
   public String toString()
   {
      String result = "";
      if ( isLeaf() )
      {
         result += element;
      }
      else
      {
         result += "(" + element;
         result += " " + ((null ==  left) ? "()" : left );  // recursion
         result += " " + ((null == right) ? "()" : right); // recursion
         result += ")";
      }
      return result;
   }
}
