/**
   This class defines a binary tree data structure,
   but it defines all of the data structure's methods
   as static methods (not as instance methods).

   We do that so that we can use the null value to
   represent an empty tree.

   In truth, this class does not define the "binary tree"
   data type. This class really defines "binary tree node"
   data type.
*/
class BTree<T>
{
   private T element;
   private BTree<T> left;
   private BTree<T> right;

   /**
      Construct a leaf node.
   */
   public BTree(T element)
   {
      this(element, null, null);
   }


   /**
      Construct a BTree with the given binary trees
      as its left and right branches.
   */
   public BTree(T element, BTree<T> left, BTree<T> right)
   {
      if ( null == element )
         throw new NullPointerException("BTree must have a root element");

      this.element = element;
      this.left  = left;
      this.right = right;
   }

   public T root()
   {
      return element;
   }

   public BTree<T> left()
   {
      return left;
   }

   public BTree<T> right()
   {
      return right;
   }


// NOTE: All of the following methods are public static methods of the BTree class.
//       There are no instance methods in this class.

   /**
      Determine if the given binary tree is a leaf node.
   */
   public static <T> boolean isLeaf(BTree<T> btree)
   {
      return null != btree
          && null == btree.left
          && null == btree.right;
   }


   /**
      Compute the maximum degree of the given BTree.
   */
   public static <T> int degree(BTree<T> btree)
   {
      if ( null == btree ) // btree is empty
      {
         return 0;
      }
      else // btree is not empty
      {
         int rootDegree = 0;

         if ( null != btree.left )
         {
            ++rootDegree;
         }
         if ( null != btree.right )
         {
            ++rootDegree;
         }
         return Math.max(rootDegree, Math.max(degree(btree.left),    // recursion
                                              degree(btree.right))); // recursion
      }
   }


   /**
      Calculate the height of the given BTree.
   */
   public static <T> int height(BTree<T> btree)
   {
      if ( null == btree ) // btree is empty
      {
         return -1;
      }
      else // btree is not empty
      {
         return 1 + Math.max(height(btree.left),   // recursion
                             height(btree.right)); // recursion
      }
   }


   /**
      Calculate the size of the given BTree.
   */
   public static <T> int size(BTree<T> btree)
   {
      if ( null == btree ) // btree is empty
      {
         return 0;
      }
      else // btree is not empty
      {
         return 1 + size(btree.left)   // recursion
                  + size(btree.right); // recursion
      }
   }


   /**
      Determine if the given BTree is a full binary tree.
   */
   public static <T> boolean isFull(BTree<T> btree)
   {
      if ( null == btree ) // btree is empty
      {
         return true;
      }
      else // btree is not empty
      {
         return ( null == btree.left && null == btree.right )
               || ( null != btree.left
                 && null != btree.right
                 && isFull( btree.left )     // recursion
                 && isFull( btree.right ) ); // recursion
      }
   }


   /**
      Do a pre-order traversal of the given binary tree.
   */
   public static <T> String preOrder(BTree<T> btree)
   {
      if ( null == btree ) // btree is empty
      {
         return "";
      }
      else
      {
         return btree.element + " "
              + preOrder(btree.left)    // recursion
              + preOrder(btree.right);  // recursion
      }
   }


   /**
      Do an in-order traversal of the given binary tree.
   */
   public static <T> String inOrder(BTree<T> btree)
   {
      if ( null == btree ) // btree is empty
      {
         return "";
      }
      else
      {
         return inOrder(btree.left)    // recursion
              + btree.element + " "
              + inOrder(btree.right);  // recursion
      }
   }


   /**
      Do a post-order traversal of the given binary tree.
   */
   public static <T> String postOrder(BTree<T> btree)
   {
      if ( null == btree ) // btree is empty
      {
         return "";
      }
      else
      {
         return postOrder(btree.left)   // recursion
              + postOrder(btree.right)  // recursion
              + btree.element + " ";
      }
   }


   /**
      Convert the given binary tree into a String.
   */
   public static <T> String toString(BTree<T> btree)
   {
      String result = "";
      if ( null == btree ) // btree is empty
      {
         result += "()";
      }
      else if ( isLeaf(btree) )
      {
         result += btree.root();
      }
      else
      {
         result += "(" + btree.root();
         result += " " + toString(btree.left());  // recursion
         result += " " + toString(btree.right()); // recursion
         result += ")";
      }
      return result;
   }
}
