/**
   This class defines a binary tree data structure
   as an object that has a reference to a binary
   tree node.

   This class gives us a well defined empty binary tree
   that is not represented by a null value. The empty
   btree is internally denoted by a null reference, but
   that reference is hidden inside of a BTree object, so
   the null reference is not part of the public interface.

   Compared to the previous definitions of the BTree class,
   this BTree "has a" node, whereas in the previous classes,
   a BTree "is a" node.

   The BTreeNode data structure is defines as a private,
   nested class inside of this BTree class.

   These trees are immutable.
*/
public class BTree<T>
{
   private BTreeNode<T> btreeNode;

   /**
      Construct an empty binary tree.
   */
   public BTree()
   {
      btreeNode = null;
   }


   /**
      Construct a leaf node.

      Notice that if this constructor didn't exist, you could
      still construct a leaf node, it would just be cumbersome.
      For example,

      BTree<String> leaf = new BTree<>("a", new BTree<>(), new BTree<>());

      So this is really a "convenience" constructor. It doesn't
      need to be defined, but it sure is convenient for it to be
      here.
   */
   public BTree(T element)
   {
      if (null == element)
         throw new IllegalArgumentException("BTree must have a root element");

      btreeNode = new BTreeNode<T>(element, null, null);
   }


   /**
      Construct a BTree with the given binary trees
      as its left and right branches.
   */
   public BTree(T element, BTree<T> left, BTree<T> right)
   {
      if (null == element)
         throw new IllegalArgumentException("BTree must have a root element");
      if (null == left)
         throw new IllegalArgumentException("BTree must have a left branch");
      if (null == right)
         throw new IllegalArgumentException("BTree must have a right branch");

      // We need to "unwrap" the nodes from the left and right branches.
      btreeNode = new BTreeNode<T>(element, left.btreeNode, right.btreeNode);
   }


   public boolean isEmpty()
   {
      return null == btreeNode;
   }


   public T root()
   {
      if (null == btreeNode)
         throw new java.util.NoSuchElementException("empty BTree");

      return btreeNode.element;
   }


   public BTree<T> left()
   {
      if (null == btreeNode)
         throw new java.util.NoSuchElementException("empty BTree");

      // We need to "wrap" the node for the
      // left branch in a BTree object.
      BTree<T> temp = new BTree<T>();
      temp.btreeNode = this.btreeNode.left;
      return temp;
   }


   public BTree<T> right()
   {
      if (null == btreeNode)
         throw new java.util.NoSuchElementException("empty BTree");

      // We need to "wrap" the node for the
      // right branch in a BTree object.
      BTree<T> temp = new BTree<T>();
      temp.btreeNode = this.btreeNode.right;
      return temp;
   }


   /**
      Determine if this binary tree is a leaf node.
   */
   public boolean isLeaf()
   {
      return null != btreeNode
          && null == btreeNode.left
          && null == btreeNode.right;
   }


   /**
      Compute the maximum degree of this BTree.
   */
   public int degree()
   {
      if ( null == btreeNode ) // this tree is empty
      {
         return 0;
      }
      else // this tree is not empty
      {
         int rootDegree = 0;

         if ( null != btreeNode.left )
         {
            ++rootDegree;
         }
         if ( null != btreeNode.right )
         {
            ++rootDegree;
         }
         return Math.max(rootDegree, Math.max( left().degree(),   // recursion
                                              right().degree())); // recursion
      }
   }


   /**
      Calculate the height of this BTree.
   */
   public int height()
   {
      if ( null == btreeNode ) // this tree is empty
      {
         return -1;
      }
      else // this tree is not empty
      {
         return 1 + Math.max( left().height(),  // recursion
                             right().height()); // recursion
      }
   }


   /**
      Calculate the size of this BTree.
   */
   public int size()
   {
      if ( null == btreeNode ) // this tree is empty
      {
         return 0;
      }
      else // this tree is not empty
      {
         return 1 +  left().size()  // recursion
                  + right().size(); // recursion
      }
   }


   /**
      Determine if this BTree is a full binary tree.
   */
   public boolean isFull()
   {
      if ( null == btreeNode ) // this tree is empty
      {
         return true;
      }
      else // this tree is not empty
      {
         return ( null == btreeNode.left && null == btreeNode.right )
               || ( null != btreeNode.left
                 && null != btreeNode.right
                 &&  left().isFull()    // recursion
                 && right().isFull() ); // recursion
      }
   }


   /**
      Do a pre-order traversal of this binary tree.
   */
   public String preOrder()
   {
      if ( null == btreeNode ) // this tree is empty
      {
         return "";
      }
      else
      {
         return btreeNode.element + " "
              +  left().preOrder()   // recursion
              + right().preOrder();  // recursion
      }
   }


   /**
      Do an in-order traversal of this binary tree.
   */
   public String inOrder()
   {
      if ( null == btreeNode ) // this tree is empty
      {
         return "";
      }
      else
      {
         return  left().inOrder()    // recursion
              + btreeNode.element + " "
              + right().inOrder();   // recursion
      }
   }


   /**
      Do a post-order traversal of this binary tree.
   */
   public String postOrder()
   {
      if ( null == btreeNode ) // this tree is empty
      {
         return "";
      }
      else
      {
         return  left().postOrder()   // recursion
              + right().postOrder()   // recursion
              + btreeNode.element + " ";
      }
   }


   /**
      Convert this binary tree into a String.
   */
   @Override
   public String toString()
   {
      String result = "";
      if ( isEmpty() )
      {
         result += "()";
      }
      else if ( isLeaf() )
      {
         result += root();
      }
      else
      {
         result += "(" + root();
         result += " " + left();  // recursion
         result += " " + right(); // recursion
         result += ")";
      }
      return result;
   }


   // A private nested class definition.
   private class BTreeNode<T>
   {
      public T element;
      public BTreeNode<T> left;
      public BTreeNode<T> right;

      public BTreeNode(T data)
      {
         this(data, null, null);
      }

      public BTreeNode(T element, BTreeNode<T> left, BTreeNode<T> right)
      {
         this.element = element;
         this.left = left;
         this.right = right;
      }

      public String toString()
      {
         if (null == left && null == right)
         {
            return element.toString();
         }
         else
         {
            String result = "(" + element;
            result += " ";
            result += (null == left) ? "()" : left;  // recursion
            result += " ";
            result += (null == right) ? "()" : right; // recursion
            result += ")";
            return result;
         }
      }
   }//BTreeNode
}
