/**
   This class defines a binary tree data structure as
   an object that has three references, a reference to
   an element, a reference to a left binary tree, and
   a reference to a right binary tree.

   This class essentially defines a binary tree as a
   binary tree node. This class solves the algebraic data
   type problem of defining the OR type by defining the
   empty tree as a node with all three references equal
   to null. So we are once again using a "tagged union"
   but now the "tag" is all three references combined.
      https://en.wikipedia.org/wiki/Tagged_union

   This class gives us a well defined empty binary tree.
   The empty btree is internally denoted by three null
   references, but those references are hidden inside of
   a BTree object, so the null references are not part of
   the public interface.
*/
public class BTree<T>
{
   private T element;
   private BTree<T> left;
   private BTree<T> right;

   /**
      Construct an empty binary tree.
   */
   public BTree()
   {
      this.element = null;
      this.left = null;
      this.right = null;
   }


   /**
      Construct a leaf node.
   */
   public BTree(T element)
   {
      if (null == element)
         throw new IllegalArgumentException("BTree must have a root element");

      this.element = element;
      this.left = null;
      this.right = null;
   }


   /**
      Construct a BTree with the given binary trees
      as its left and right branches.
   */
   public BTree(T element, BTree<T> left, BTree<T> right)
   {
      if (null == element)
         throw new IllegalArgumentException("BTree must have a root element");
      if (null == left)
         throw new IllegalArgumentException("BTree must have a left branch");
      if (null == right)
         throw new IllegalArgumentException("BTree must have a right branch");

      this.element = element;
      this.left  = left;
      this.right = right;
   }


   public boolean isEmpty()
   {
      return null == element
          && null == left
          && null == right;
   }


   public T root()
   {
      if ( isEmpty() )
         throw new java.util.NoSuchElementException("empty BTree");

      return element;
   }


   public BTree<T> left()
   {
      if ( isEmpty() )
      {
         throw new java.util.NoSuchElementException("empty BTree");
      }
      else if ( null == left )
      {
         return new BTree<T>(); // return an empty tree
      }
      else
      {
         return left;
      }
   }


   public BTree<T> right()
   {
      if ( isEmpty() )
      {
         throw new java.util.NoSuchElementException("empty BTree");
      }
      else if ( null == right )
      {
         return new BTree<T>(); // return an empty tree
      }
      else
      {
         return right;
      }
   }


   /**
      Determine if this binary tree is a leaf node.
   */
   public boolean isLeaf()
   {
      return null != element
          && null == left
          && null == right;
   }


   /**
      Compute the maximum degree of this BTree.
   */
   public int degree()
   {
      if ( isEmpty() )
      {
         return 0;
      }
      else // this tree is not empty
      {
         int rootDegree = 0;

         if ( null != left )
         {
            ++rootDegree;
         }
         if ( null != right )
         {
            ++rootDegree;
         }
         return Math.max(rootDegree, Math.max( left().degree(),   // recursion
                                              right().degree())); // recursion
      }
   }


   /**
      Calculate the height of this BTree.
   */
   public int height()
   {
      if ( isEmpty() )
      {
         return -1;
      }
      else // this tree is not empty
      {
         return 1 + Math.max( left().height(),  // recursion
                             right().height()); // recursion
      }
   }


   /**
      Calculate the size of this BTree.
   */
   public int size()
   {
      if ( isEmpty() )
      {
         return 0;
      }
      else // this tree is not empty
      {
         return 1 +  left().size()  // recursion
                  + right().size(); // recursion
      }
   }


   /**
      Determine if this BTree is a full binary tree.
   */
   public boolean isFull()
   {
      if ( isEmpty() )
      {
         return true;
      }
      else // this tree is not empty
      {
         return ( isLeaf() )
               || ( !  left().isEmpty()
                 && ! right().isEmpty()
                 &&  left().isFull()    // recursion
                 && right().isFull() ); // recursion
      }
   }


   /**
      Do a pre-order traversal of this binary tree.
   */
   public String preOrder()
   {
      if ( isEmpty() )
      {
         return "";
      }
      else
      {
         return element + " "
              +  left().preOrder()   // recursion
              + right().preOrder();  // recursion
      }
   }


   /**
      Do an in-order traversal of this binary tree.
   */
   public String inOrder()
   {
      if ( isEmpty() )
      {
         return "";
      }
      else
      {
         return  left().inOrder()    // recursion
              + element + " "
              + right().inOrder();   // recursion
      }
   }


   /**
      Do a post-order traversal of this binary tree.
   */
   public String postOrder()
   {
      if ( isEmpty() )
      {
         return "";
      }
      else
      {
         return  left().postOrder()  // recursion
              + right().postOrder()  // recursion
              + element + " ";
      }
   }


   /**
      Convert this binary tree into a String.
   */
   @Override
   public String toString()
   {
      String result = "";
      if ( isEmpty() )
      {
         result += "()";
      }
      else if ( isLeaf() )
      {
         result += root();
      }
      else
      {
         result += "(" + root();
         result += " " + left();  // recursion
         result += " " + right(); // recursion
         result += ")";
      }
      return result;
   }
}
