/**
   This is an abstract base class for defining a binary
   tree data type.

   We will define two concrete subclasses of this base
   class. One subclass will define binary tree nodes.
   The other subclass will define the empty binary tree.
   The two subclasses give us the OR (the "sum" type)
   in the algebraic data type for a binary tree.

   Notice that even though this is an abstract class,
   the interface that it defines is sufficient to write
   concrete implementations of all the recursive methods
   we need on binary trees.
*/
public abstract class BTree<T>
{
   public abstract boolean isEmpty();

   public abstract T root();

   public abstract BTree<T> left();

   public abstract BTree<T> right();


   // Use the above four (abstract) methods to define
   // all the recursive methods we need on binary trees.

   /**
      Determine if this binary tree is a leaf node.
   */
   public boolean isLeaf()
   {
      return ! isEmpty()
          && left().isEmpty()
          && right().isEmpty();
   }


   /**
      Compute the maximum degree of this BTree.
   */
   public int degree()
   {
      if ( isEmpty() )
      {
         return 0;
      }
      else // non empty tree
      {
         int rootDegree = 0;

         if ( ! left().isEmpty() )
         {
            ++rootDegree;
         }
         if ( ! right().isEmpty() )
         {
            ++rootDegree;
         }
         return Math.max(rootDegree, Math.max( left().degree(),  // recursion
                                              right().degree())); // recursion
      }
   }


   /**
      Calculate the height of this BTree.
   */
   public int height()
   {
      if ( isEmpty() )
      {
         return -1;
      }
      else // non empty tree
      {
         return 1 + Math.max( left().height(),  // recursion
                             right().height()); // recursion
      }
   }


   /**
      Calculate the size of this BTree.
   */
   public int size()
   {
      if ( isEmpty() )
      {
         return 0;
      }
      else // non empty tree
      {
         return 1 +  left().size()  // recursion
                  + right().size(); // recursion
      }
   }


   /**
      Determine if this BTree is a full binary tree.
   */
   public boolean isFull()
   {
      if ( isEmpty() )
      {
         return true;
      }
      else // non empty tree
      {
         return ( left().isEmpty() && right().isEmpty() )
               || ( !  left().isEmpty()
                 && ! right().isEmpty()
                 &&  left().isFull()    // recursion
                 && right().isFull() ); // recursion
      }
   }


   /**
      Do a pre-order traversal of this binary tree.
   */
   public String preOrder()
   {
      if ( isEmpty() )
      {
         return "";
      }
      else
      {
         return root() + " "
              +  left().preOrder()  // recursion
              + right().preOrder(); // recursion
      }
   }


   /**
      Do an in-order traversal of this binary tree.
   */
   public String inOrder()
   {
      if ( isEmpty() )
      {
         return "";
      }
      else
      {
         return  left().inOrder()   // recursion
              + root() + " "
              + right().inOrder();  // recursion
      }
   }


   /**
      Do a post-order traversal of this binary tree.
   */
   public String postOrder()
   {
      if ( isEmpty() )
      {
         return "";
      }
      else
      {
         return  left().postOrder()  // recursion
              + right().postOrder()  // recursion
              + root() + " ";
      }
   }


   /**
      Convert this binary tree into a String.
   */
   @Override
   public String toString()
   {
      String result = "";
      if ( isEmpty() )
      {
         result += "()";
      }
      else if ( isLeaf() )
      {
         result += root();
      }
      else
      {
         result += "(" + root();
         result += " " + left();  // recursion
         result += " " + right(); // recursion
         result += ")";
      }
      return result;
   }

}
