/**
   This class defines a binary tree where every node
   always has exactly two branches. This definition
   of a binary tree means that there is no such thing
   as an empty tree. This simplifies the definitions
   of all the recursive methods in this class.

   Notice that this class really is a definition of a
   binary tree node. But both cases in the algebraic
   data type definition for this binary tree can be
   implemented by a node. The "tag" that tells us
   which kind of tree we have is the combination of
   the two child references. If both of them are null,
   then this is a leaf node.
*/
public class BTree<T>
{
   private T element;
   private BTree<T> left;
   private BTree<T> right;

   /**
      Construct a leaf node.
   */
   public BTree(T element)
   {
      if (null == element)
         throw new IllegalArgumentException("BTree must have a root element");

      this.element = element;
      this.left = null;
      this.right = null;
   }


   /**
      Construct a BTree with the given binary trees
      as its left and right branches.
   */
   public BTree(T element, BTree<T> left, BTree<T> right)
   {
      if (null == element)
         throw new IllegalArgumentException("BTree must have a root element");
      if (null == left)
         throw new IllegalArgumentException("BTree must have a left branch");
      if (null == right)
         throw new IllegalArgumentException("BTree must have a right branch");

      this.element = element;
      this.left  = left;
      this.right = right;
   }

   public boolean isEmpty()
   {
      return false;
   }

   public T root()
   {
      return element;
   }


   public BTree<T> left()
   {
      return left;
   }


   public BTree<T> right()
   {
      return right;
   }


   /**
      Determine if this binary tree is a leaf node.
   */
   public boolean isLeaf()
   {
      return null == left
          && null == right;
   }


   /**
      Compute the maximum degree of this BTree.
   */
   public int degree()
   {
      if ( isLeaf() )
      {
         return 0;
      }
      else
      {
         return 2;
      }
   }


   /**
      Calculate the height of this BTree.
   */
   public int height()
   {
      if ( isLeaf() )
      {
         return 0;
      }
      else
      {
         return 1 + Math.max( left.height(),  // recursion
                             right.height()); // recursion
      }
   }


   /**
      Calculate the size of this BTree.
   */
   public int size()
   {
      if ( isLeaf() )
      {
         return 1;
      }
      else
      {
         return 1 +  left.size()  // recursion
                  + right.size(); // recursion
      }
   }


   /**
      Determine if this BTree is a full binary tree.
   */
   public boolean isFull()
   {
      return true;
   }


   /**
      Do a pre-order traversal of this binary tree.
   */
   public String preOrder()
   {
      if (  isLeaf() )
      {
         return element + " ";
      }
      else
      {
         return element + " "
              +  left.preOrder()  // recursion
              + right.preOrder(); // recursion
      }
   }


   /**
      Do an in-order traversal of this binary tree.
   */
   public String inOrder()
   {
      if (  isLeaf() )
      {
         return element + " ";
      }
      else
      {
         return left.inOrder()   // recursion
              + element + " "
              + right.inOrder(); // recursion
      }
   }


   /**
      Do a post-order traversal of this binary tree.
   */
   public String postOrder()
   {
      if (  isLeaf() )
      {
         return element + " ";
      }
      else
      {
         return  left.postOrder()  // recursion
              + right.postOrder()  // recursion
              + element + " ";
      }
   }


   /**
      Convert this binary tree into a String.
   */
   @Override
   public String toString()
   {
      String result = "";
      if ( isLeaf() )
      {
         result += element;
      }
      else
      {
         result += "(" + element;
         result += " " + left;   // recursion
         result += " " + right;  // recursion
         result += ")";
      }
      return result;
   }
}
